#include <avr/interrupt.h>

#include <util/delay.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <inttypes.h>

#include <stdlib.h>

#define F_SMP		(32000UL)		/* sampling frequency */
#define INT_CYCLES	(F_CPU / F_SMP)	/* interrupt period in timer cycles */
#define LOWSMPR_FAC	3				/* low sample rate divider 32k/(1<<3) = 4kHz */

#define N_BANDS		12				/* number of bands */

#define BARS_SPEED	24	/* bars falling down speed = BARS_FALL * 6 dB/s */
#define PEAK_DELAY	200	/* time after wich peaks start to fall down = PEAK_DELAY * 4.096ms*/
#define PEAK_SPEED	6	/* peaks falling down speed = PEAK_FALL * 6 dB/s */

#define F_DISPLAY_INT	(F_CPU/64/256)	/* display (timer0) interrupt frequency */

#define DIV_BFALL	(F_DISPLAY_INT/BARS_SPEED)
#define DIV_PFALL	(F_DISPLAY_INT/PEAK_SPEED)
#define DIV_PDEL	(F_DISPLAY_INT/200)

/* return greater number of two given */
#define max(a, b)	(((a)>(b)) ? (a) : (b))

/* fft window gain = WINDOW_GAIN * (-6) dB (just an approximation) */
#define WINDOW_GAIN 1

/* Defs for using Fixed-point FFT module */
#include "ffft.h"

#include "bars_array_8.h"

typedef struct _tag_complex_t {
	int16_t r;
	int16_t i;
} complex_t;

void fft_input (const int16_t *, complex_t *);
void fft_execute (complex_t *);
void fft_output (complex_t *, uint16_t *);
int16_t fmuls_f (int16_t, int16_t);

int16_t capture[FFT_N];	/* Wave capture buffer */
complex_t bfly_buff[FFT_N];			/* FFT buffer */
uint16_t spectrum[FFT_N/2];		/* Spectrum output buffer */

volatile uint16_t bars[N_BANDS];	/* packed linear spectrum value */
volatile uint8_t bands[N_BANDS];	/* bands bar height */
volatile uint8_t peaks[N_BANDS];	/* peaks position */
volatile uint8_t peak_delay[N_BANDS];	/* peaks fall down delay */
volatile uint8_t leds[N_BANDS];	/* LED matrix image */

#define ADMUX_SH (_BV(ADLAR) + 0)	/* ADMUX register value when sampling with full rate */
#define ADMUX_SL (_BV(ADLAR) + 1)	/* ADMUX register value when sampling with low rate */

#define BUFF_FILL	0				/* buffer is being filled */
#define RATE_LOW	1				/* select low sample rate */
#define PEAKS_ON	2				/* peak amplitude indicators */
#define	BARS_FALL	3
volatile uint8_t flags=_BV(PEAKS_ON) | _BV(BARS_FALL);

/*
** values used for controlling column selector circuit
** format (MSb first): xxLHnnnn
** x - don't care
** L, H - turn on lower/higher respectively half of columns
** n - binary coded column number (0-9), MSb first
*/
static const PROGMEM uint8_t column_ctrl[] = {
	0x20, 0x21, 0x22, 0x23, 0x24, 0x25,
	0x10, 0x11, 0x12, 0x13, 0x14, 0x15
	};


/**************************************************************************/

/* interrupt handler called each time timer0 overflows */

ISR(SIG_OVERFLOW0) {
static uint16_t div_bfall=DIV_BFALL;
static uint16_t div_pfall=DIV_PFALL;
static uint8_t div_pdel=DIV_PDEL;
static uint8_t column=N_BANDS-1;
uint8_t bar_num;

	PORTC |= (uint8_t) _BV(PC2);

/* signal magnitude indicators fall down */
	if(flags & _BV(BARS_FALL)) {
		if((--div_bfall) == 0) {
			div_bfall = DIV_BFALL;
			for(bar_num=0; bar_num<N_BANDS; bar_num++)
				if(bands[bar_num])
					bands[bar_num] -= 1;
		}
	}

	if(flags & _BV(PEAKS_ON)) {
/* delay after wich peaks will start to fall down */
		if((--div_pdel) == 0) {
			div_pdel = DIV_PDEL;
			for(bar_num=0; bar_num<N_BANDS; bar_num++)
				if(peak_delay[bar_num])
					peak_delay[bar_num] -= 1;
		}

/* peak indicators fall down */
		if((--div_pfall) == 0) {
			div_pfall = DIV_PFALL;
			for(bar_num=0; bar_num<N_BANDS; bar_num++)
/*				if((peak_delay[bar_num] == 0) && (peaks[bar_num] != 0))*/
				if(peak_delay[bar_num]==0)
					if(peaks[bar_num])
						peaks[bar_num] -= 1;
		}
	}

/* Refresh display */
	if(column-- == 0)
		column = N_BANDS-1;
		
	PORTD = 0xff;		/* blank LEDs */
	PORTB = pgm_read_byte_near(column_ctrl + column);	/* switch column */
	PORTD = leds[column];

	PORTC &= (uint8_t) (~_BV(PC2));

}

/* interrupt handler called every timer1 compare match */

ISR(SIG_OUTPUT_COMPARE1A) {
static	uint8_t bufi=0;				/* input buffer index */
static uint8_t lowsmpr_cnt=(1<<LOWSMPR_FAC);
int16_t adc_smp;

	PORTC |= _BV(PC2);

	adc_smp = ((int16_t) ADC) - 0x8000;	/* read result of previous conversion */
	ADCSRA |= _BV(ADSC);					/* start new conversion */
	if(flags & _BV(BUFF_FILL)) {
		if(flags & _BV(RATE_LOW)) {
			if((--lowsmpr_cnt) == 0) {
				capture[bufi++] = adc_smp;
				lowsmpr_cnt = (1<<LOWSMPR_FAC);
			}
		} else {
			capture[bufi++] = adc_smp;	/* sampling with full rate */
		}

		if(bufi == FFT_N) {						/* buffer filled */
			bufi = 0;
			flags &= (uint8_t) (~_BV(BUFF_FILL));	/* clear flag */
		}
	}
	

/* overdrive indicator */
	if((((adc_smp & 0xff00) >> 8) < -120) || (((adc_smp & 0xff00) >> 8) > 120)) {
		PORTC |= _BV(PC3);
	} else {
		PORTC &= ~_BV(PC3);
	}

	PORTC &= (~_BV(PC2)) & 0xff;
}

/**************************************************************************/

void init(void) {

	PORTD = 0xff;
	DDRD = 0xff;

	PORTB = 0x3f;
	DDRB = 0x3f;

/*
** PC2 - cpu usage (PWM)
** PC3 - overdrive indicator (LED)
** PC4 - high when computing FFT
** PC5 - pulse every complete cycle
*/

	PORTC = 0;
	DDRC = _BV(PC2) | _BV(PC3) | _BV(PC4) | _BV(PC5);


/*
** enable ADC, prescaler = 16 -> ADC clock = 16MHz / 32 = 500kHz
** ADC is overclocked
*/
	ADCSRA = _BV(ADEN) + 5;
	

/* chanel 0, external reference voltage, left adjusted result */
	ADMUX = ADMUX_SH;
	
/* Start a dummy conversion, and wait for result */
	ADCSRA |= _BV(ADSC);
	

/* configure timer 1 & 0 */

/* Timer 1: CTC mode TOP value = OCR1A, no prescaler */
	TCCR1A = 0;	
	TCCR1B = _BV(CS10) | _BV(WGM12);

/* Timer 0: prescaler = clk_IO / 64 */
	TCCR0 = _BV(CS01) | _BV(CS00);

/* Timer 1: interrupt every INT_CYCLES cycles */
	OCR1A = INT_CYCLES;

/* Timer 1: enable interrupt on output compare A match */
/* Timer 0: enable interrupt on overflow */
	TIMSK = _BV(OCIE1A) | _BV(TOIE0);

	//sei();					/* enable interupts */

}

/**************************************************************************/
/* count leading zeros */

uint8_t clz(uint16_t x) { 
uint8_t n;
	x = (uint16_t) (((uint32_t) x * x) >> 16);
	if (x == 0) return(16);
	n = 0;
	
	while((x & 0x8000) == 0) {
		n++;
		x <<=1;
	}

	return(n);
}

/**************************************************************************/
/* integer base 2 logarithm */

uint8_t ilog2(uint16_t x) { 
uint8_t n;
	if (x == 0) return(16);
	n = 0;
	
	while((x & 0x8000) == 0) {
		n++;
		x <<=1;
	}

	return(n);
}


/**************************************************************************/

void pack_spectrum(const uint8_t *window,
					uint8_t shift, uint16_t nbands) {
uint8_t band, bin;
uint16_t sig;
uint16_t *spectr_mag = spectrum;

	for(band=shift; band<(nbands+shift); band++) {
		sig = 0;
		for(bin=0; bin<64; bin++) {
			sig += ((uint32_t) ((*++spectr_mag) & 0x0000ffff) * (((uint32_t) pgm_read_byte_near(window++) & 0x000000ff))) >> 8;
		}
		spectr_mag -= 64;
		bars[band] = sig;
	}
}

/**************************************************************************/

void make_image(void) {
uint8_t band, bar_height, led;
uint16_t sig;

	for(band = 0; band < N_BANDS; band++) {

		sig = bars[band];

		bar_height = 16 - clz(sig) + WINDOW_GAIN;
		if((bands[band] < bar_height) || ((flags & _BV(BARS_FALL)) == 0))
			bands[band] = bar_height;
		
		if((peaks[band] < bar_height) && ((flags & _BV(PEAKS_ON)) != 0)) {
			peaks[band] = bar_height;
			peak_delay[band] = PEAK_DELAY;
		}
/* create bar from height value */
		bar_height = 16;
		led = 0xff;
		while(bar_height-- > bands[band]) {
			led >>=1;
		}
/* invert, "0" means LED is on */				
		leds[band] = (uint8_t) ~led;
/* add peaks to display */
		if(flags & _BV(PEAKS_ON)) {
			leds[band] &= (uint8_t) ~((1 << (peaks[band] - 8)) >> 1);
		}
	}
}

/**************************************************************************/

int main(void) {
volatile int8_t adc_smp, samp;
volatile uint16_t danaac,dana1,dat;
volatile aaa=0xaa;
volatile uint64_t testowa=0x12345678;
volatile ppp=0xff;
volatile uint8_t tavv[8]={1,2,3,4,5,6,7,8};


fft_input(capture, bfly_buff);


fft_execute(bfly_buff);


fft_output(bfly_buff, spectrum);
testowa=0x78ffffffffffffff;
tavv[0]=testowa;
for(dat=64; dat>32; dat--)
{ testowa>>=1;
}

while(1){
}
	return 0;
}
